#! /usr/bin/env python3
import mysql.connector as mc
import time
import logging
from systemd.journal import JournalHandler

log = logging.getLogger('edgility-db')
log.addHandler(JournalHandler())
log.setLevel(logging.INFO)

class EdgilityDatabase(object):

    MAX_CONNECTION_RETRIES = 2
    CONNECTION_RETRY_DELAY = 1

    # TODO: HIDE CREDENTIALS
    def __init__(self,  user='root', password='root', database='wireguard', host='localhost'):
        self._user = user
        self._password = password
        self._database = database
        self._host = host
        self._connection = None
        self._cursor = None

    def __del__(self):
        if self._connection and self._connection.is_connected():
            self.close_cursor_and_connection()

    # For cases where cursor is used as an iterable
    @property
    def cursor(self):
        return self._cursor

    # TODO: HIDE CREDENTIALS
    def _connect_to_mysql(self):
        for retry in range(self.MAX_CONNECTION_RETRIES):
            try:
                return mc.connect(host=self._host, user=self._user, passwd=self._password, database=self._database)
            except Exception as ex:
                log.info(ex)
                time.sleep(retry * self.CONNECTION_RETRY_DELAY)

        return None
    
    def _setup_connection_and_cursor(self, buffered_cursor):
        if not self._connection:
            self._connection = self._connect_to_mysql()
        if self._connection and self._connection.is_connected():
            if not self._cursor:
                self._cursor = self._connection.cursor(buffered_cursor)

        return (self._connection and self._cursor)

    # Remember to call close_cursor_and_connection()
    def execute(self, operation, params=None, multi=False, buffered_cursor=False, auto_commit=False):
        if self._setup_connection_and_cursor(buffered_cursor):
            self._cursor.execute(operation, params, multi)
            if auto_commit:
                self._connection.commit()
            return True

        return False

    # Remember to call close_cursor_and_connection()
    def execute_many(self, operation, seq_of_params, buffered_cursor=False, auto_commit=False):
        if self._setup_connection_and_cursor(buffered_cursor):
            self._cursor.execute_many(operation, seq_of_params)
            if auto_commit:
                self._connection.commit()
            return True

        return False

    # Suitable for single database select query
    def execute_fetch_and_close(self, operation, params=None, multi=False, fetch_one=False):
        data = []
        if self.execute(operation, params, multi):
            if fetch_one:
                data = self._cursor.fetchone()
            else:
                data = self._cursor.fetchall()
                data = [item[0] for item in data] if type(data) is tuple else data

            self.close_cursor_and_connection()

        return data

    # Suitible for single insert query
    def execute_and_close(self, operation, params=None, multi=False, auto_commit=True):
        try:
            if self.execute(operation, params, multi, auto_commit=auto_commit):
                log.error(f"Query executed successfully: {operation}")
            else:
                log.error(f"Failed to execute query: {operation}")
        except Exception as e:
            log.error(f"Error executing query: {operation}. Error: {e}")
        finally:
            if self._cursor:
                self.close_cursor_and_connection()
            else:
                log.debug("Cursor is None, cannot close cursor and connection.")
    
    def create_database(self):
        if not self._setup_connection_and_cursor(False):
            self._cursor.execute(f'CREATE DATABASE {self._database}')

    # If execute() is called without auto_commit enabled
    def commit(self):
        if self._connection and self._connection.is_connected():
            self._connection.commit()

    def close_cursor(self):
        self._cursor.close()
        self._cursor = None

    def close_connection(self):
        self._connection.close()
        self._connection = None

    def close_cursor_and_connection(self):
        self.close_cursor()
        self.close_connection()
    
    def create_table(self, table_name, fields):
        create_query = f"CREATE TABLE IF NOT EXISTS {table_name} ({fields});"
        self.execute_and_close(create_query, auto_commit=True)
        
    def drop_table(self, table_name):
        create_query = f"DROP TABLE {table_name};"
        self.execute_and_close(create_query, auto_commit=True)

    def get_all_tables(self, database_name):
        query = f"SELECT table_name FROM information_schema.tables \
        WHERE table_schema = '{database_name}'"
        
        results = self.execute_fetch_and_close(query)
        # table_names = [result[0] for result in results if isinstance(result[0], str)]
        table_names = [result[0] for result in results if isinstance(result, tuple) and len(result) == 1 and isinstance(result[0], str)]
        log.error(f"all tables {table_names}")
        return table_names

    def insert_values_in_table(self, table_name, fields, values):
        insert_query = f"INSERT INTO {table_name} ({', '.join(fields)}) VALUES ({'%s,'*(len(fields)-1)}%s);"
        log.error(f'PESHO IVO {insert_query}')
        self.execute_and_close(insert_query, values, auto_commit=True)

    def delete_value_from_table(self, table_name, value, data):
        delete_query = f"DELETE FROM {table_name} WHERE {value} = '{data}';"
        self.execute_and_close(delete_query, table_name, auto_commit=True)

    def fetch_from_table_by_param(self, table_name, column, param, data):
        fetch_query = f"SELECT {column} FROM {table_name} WHERE {param} = '{data}';"
        result = self.execute_fetch_and_close(fetch_query, table_name, fetch_one=True)
        return result[0] if result else None
    
    def fetch_column_entries(self, table_name, column):
        fetch_query = f"SELECT {column} FROM {table_name};"
        result = self.execute_fetch_and_close(fetch_query, fetch_one=False)
        return result if result else None

    def fetch_row_by_param(self, table_name, param, value):
        fetch_query = f'SELECT * FROM {table_name} WHERE {param} = "{value}";'
        result = self.execute_fetch_and_close(fetch_query, fetch_one=True)
        return result if result else None

    def fetch_all_rows_from_table(self, table_name):
        fetch_query = f'SELECT * FROM {table_name};'
        result = self.execute_fetch_and_close(fetch_query)
        return result if result else []
    
    def verify_if_entry_exists_in_table_by_param(self, table_name, param, value):
        verify_query = f'SELECT EXISTS(SELECT * FROM {table_name} WHERE {param} = "{value}");'
        result = self.execute_fetch_and_close(verify_query)
        if result:
            return True if result[0][0] == 1 else False
        else:
            return False
    
    def update_column_value_on_matching_key(self, table_name, column, value, key, key_value):
        update_query = f'Update {table_name} SET {column} = {value} WHERE {key} = "{key_value}";'
        self.execute_and_close(update_query, auto_commit=True)
        
    def insert_values_if_not_exist_otherwise_update(self, table_name, fields, values, column, value_to_update):
        insert_query = f"INSERT INTO {table_name} ({', '.join(fields)}) VALUES {values} ON DUPLICATE KEY UPDATE {column}={value_to_update};"
        # log.error(f"new query {insert_query}")
        self.execute_and_close(insert_query, auto_commit=True)

    def grant_privileges_to_user(self, username, database, table, privileges, host='%'):
        grant_query = f"GRANT {', '.join(privileges)} ON `{database}`.`{table}` TO '{username}'@'{host}';"
        self.execute_and_close(grant_query, auto_commit=True)

    def user_exists(self, username):
        query = f"SELECT COUNT(*) FROM mysql.user WHERE user = '{username}';"
        result = self.execute_fetch_and_close(query, fetch_one=True)
        return result[0] > 0

    def create_user(self, username, password):
        if self.user_exists(username):
            log.info(f"User '{username}' already exists. Skipping user creation.")
        else:
            create_user_query = f"CREATE USER '{username}' IDENTIFIED BY '{password}';"
            self.execute_and_close(create_user_query, auto_commit=True)
            log.info(f"User '{username}' created successfully.")
